from .RealInterval import ClassicalArithmetic, KaucherArithmetic, ArrayInterval, precision, SingleInterval, Interval, ARITHMETICS, INTERVAL_CLASSES, single_type
from .utils import *
from .visualization import lineqs, IntLinIncR2, lineqs3D, IntLinIncR3, OneShotVisual2D, scatter_plot

from .linear import *
from .nonlinear import *
